// Драйвер для PCI платы

#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/kdev_t.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <asm/uaccess.h>
#include <linux/interrupt.h>
#include <linux/pagemap.h>
#include <linux/sched.h>

#include "nmx.h"
#include "nmxioctl.h"

MODULE_LICENSE("Dual BSD/GPL");

static atomic_t Boards_count = ATOMIC_INIT(0); // Количество плат
static int Major; // Выбранный старший номер устройства
static struct class *nmx_class; // Класс

// Список устройств, поддерживаемых драйвером
static struct pci_device_id nmx_ids[] =
{
	{PCI_DEVICE(VENDOR_ID, DEVICE_ID),},
	{0, }
};

// Экспортирование структуры ids в пользовательское пространство
MODULE_DEVICE_TABLE(pci, nmx_ids);

static int nmx_probe(struct pci_dev *dev, const struct pci_device_id *id);
static void nmx_remove(struct pci_dev *dev);
static int nmx_init_device(struct nmx_device_extension *dev_ex);
static void nmx_deinit_device(struct nmx_device_extension *dev_ex);
static int nmx_open(struct inode *inode, struct file *filp);
static int nmx_release(struct inode *inode, struct file *filp);
static loff_t nmx_llseek(struct file *filp, loff_t offset, int from);
static ssize_t nmx_read(struct file *filp, char __user *buff, size_t count, loff_t *offp);
static ssize_t nmx_write(struct file *filp, const char __user *buff, size_t count, loff_t *offp);
int nmx_DMA(struct nmx_device_extension *dev_ex, unsigned long User_address, u32 Board_address, u32 Count, int rw);
int nmx_DMA_(struct nmx_device_extension *dev_ex, struct page **pages, unsigned int npages,
	unsigned long User_address, u32 Board_address, u32 Count, int rw);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
static long nmx_ioctl(struct file *file, unsigned int ioctl_num, unsigned long ioctl_param);
#else
static int nmx_ioctl(struct inode *inode, struct file *file, unsigned int ioctl_num, unsigned long ioctl_param);
#endif
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 18)
static irqreturn_t nmx_irq_handler(int irq, void *dev_id);
#else
static irqreturn_t nmx_irq_handler(int irq, void *dev_id, struct pt_regs *sp);
#endif
void nmx_do_tasklet(unsigned long num);

// Основная структура драйвера
static struct pci_driver pci_driver =
{
	.name = DEVICE_NAME, // Имя драйвера
	.id_table = nmx_ids, // Указатель на список устройств, поддерживаемых драйвером
	.probe = nmx_probe, // Зондирующая функция
	.remove = nmx_remove, // Вызывается, когда устройство удаляется
};

// Структура файловых операций
struct file_operations fops =
{
	.owner          = THIS_MODULE,
	.llseek         = nmx_llseek,
	.read           = nmx_read,
	.write          = nmx_write,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
	.unlocked_ioctl = nmx_ioctl,
#else
	.ioctl          = nmx_ioctl,
#endif
	.open           = nmx_open,
	.release        = nmx_release,
};

// Функция инициализации
int __init nmx_init(void)
{
	dev_t dev_numbers; // Старший и младший номера устройств
	int Return_value;

	// Создание класса
	nmx_class = class_create(THIS_MODULE, DEVICE_NAME);
	if (IS_ERR(nmx_class))
	{
		printk(KERN_ERR DEVICE_NAME": ERROR: Can not create device class!\n");
		return -ENXIO;
	}

	// Получение (динамически) номеров устройств (chrdev region)
	Return_value = alloc_chrdev_region(&dev_numbers, 0, MAX_BOARDS_COUNT, DEVICE_NAME);
	if (Return_value)
	{
		printk(KERN_ERR DEVICE_NAME": ERROR: Can not register chrdev region!\n");
		goto Fail_nmx_init_1;
	}

	// Сохранение старшего номера устройства в глобальной переменной
	Major = MAJOR(dev_numbers);

	// Регистрация основной структуры драйвера
	Return_value = pci_register_driver(&pci_driver);
	if (Return_value)
	{
		printk(KERN_ERR DEVICE_NAME": ERROR: Can not register driver!\n");
		goto Fail_nmx_init_2;
	}

	printk(KERN_INFO DEVICE_NAME": %u board(s) detected.\n", atomic_read(&Boards_count));

	// Если не обнаружено ни одной платы
	if (!atomic_read(&Boards_count))
	{
		printk(KERN_ERR DEVICE_NAME": ERROR: Board(s) not found!\n");
		Return_value = -ENODEV; // Возвращаемое значение - "Нет такого устройства"
		goto Fail_nmx_init_3;
	}

	return SUCCESS; // Нормальное завершение

Fail_nmx_init_3:

	// Разрегистрирование основной структуры драйвера
	pci_unregister_driver(&pci_driver);

Fail_nmx_init_2:

	// Освобождение номеров устройств
	unregister_chrdev_region(dev_numbers, MAX_BOARDS_COUNT);

Fail_nmx_init_1:

	// Удаление класса
	class_destroy(nmx_class);

	return Return_value; // Ошибочное завершение
}

// Функция зачистки
void __exit nmx_exit(void)
{
	dev_t dev_numbers = MKDEV(Major, 0); // Старший и младший номера устройств

	// Разрегистрирование основной структуры драйвера
	pci_unregister_driver(&pci_driver);

	// Освобождение номеров устройств
	unregister_chrdev_region(dev_numbers, MAX_BOARDS_COUNT);

	// Удаление класса
	class_destroy(nmx_class);
}

// Зондирующая функция
static int nmx_probe(struct pci_dev *dev, const struct pci_device_id *id)
{
	struct nmx_device_extension *dev_ex; // Указатель на структуру расширения текущего устройства
	dev_t dev_number; // Старший и младший номера текущего устройства
	unsigned int index = 0; // Номер устройства
	struct device *Return_value_device;
	int Return_value;

	// Получение номера текущего устройства
	index = atomic_inc_return(&Boards_count) - 1;

	dev_number = MKDEV(Major, index); // Старший и младший номера текущего устройства

	// Разрешение устройства
	Return_value = pci_enable_device(dev);
	if (Return_value)
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Can not enable device!\n", index);
		return Return_value;
	}

	// Выделение памяти (с обнолением) для структуры расширения текущего устройства
	dev_ex = kzalloc(sizeof(struct nmx_device_extension), GFP_KERNEL);
	if (!dev_ex)
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Insufficient memory!\n", index);
		Return_value = -ENOMEM;
		goto Fail_nmx_probe_1;
	}

	// Заполнение полей структуры расширения текущего устройства
	dev_ex->dev = dev; // Указатель на основную структуру устройства
	dev_ex->Num = index; // Номер устройства

	// Инициализация головы очереди ожидания
	init_waitqueue_head(&dev_ex->queue);

	// Инициализация спин-блокировки доступа к списку ожидания прерывания
	spin_lock_init(&dev_ex->Wait_interrupt_lock);

	// Инициализация списка ожидания прерывания
	INIT_LIST_HEAD(&dev_ex->Wait_interrupt_list);

	// Инициализация тасклета прерывания PCI
	tasklet_init(&dev_ex->Interrupt_tasklet, nmx_do_tasklet, (unsigned long)dev_ex);

	// Инициализация платы
	Return_value = nmx_init_device(dev_ex);
	if (Return_value)
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Can not init device!\n", index);
		goto Fail_nmx_probe_2;
	}

	// Инициализация структуры cdev
	cdev_init(&dev_ex->cdev, &fops);
	dev_ex->cdev.owner = THIS_MODULE;

	// Сообщение ядру о создании структуры cdev
	Return_value = cdev_add(&dev_ex->cdev, dev_number, 1);
	if (Return_value)
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Can not adding char device!\n", index);
		goto Fail_nmx_probe_3;
	}

	// Добавление устройства в класс (для создания узла устройства в директории /dev/)
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 18)
	Return_value_device = device_create(nmx_class, NULL, dev_number, NULL, "%s_%2.2u", DEVICE_NAME, index);
#else
	Return_value_device = device_create(nmx_class, NULL, dev_number, "%s_%2.2u", DEVICE_NAME, index);
#endif
	if (IS_ERR(Return_value_device))
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Can not adding device to class!\n", index);
		Return_value = -ENXIO;
		goto Fail_nmx_probe_4;
	}

	// Сохранение указателя на структуру расширения текущего устройства в основной структуре устройства
	pci_set_drvdata(dev, dev_ex);

	return SUCCESS; // Нормальное завершение

Fail_nmx_probe_4:

	// Удаление символьного устройства из системы
	cdev_del(&dev_ex->cdev);

Fail_nmx_probe_3:

	// Деинициализация платы
	nmx_deinit_device(dev_ex);

Fail_nmx_probe_2:

	// Запрет планирования тасклета
	tasklet_kill(&dev_ex->Interrupt_tasklet);

	// Освобождение памяти из под структуры расширения текущего устройства
	kfree(dev_ex);

Fail_nmx_probe_1:

	// Отключение устройства
	pci_disable_device(dev);

	return Return_value; // Ошибочное завершение
}

// Вызывается, когда устройство удаляется
static void nmx_remove(struct pci_dev *dev)
{
	struct nmx_device_extension *dev_ex; // Указатель на структуру расширения текущего устройства

	// Получение указателя на структуру расширения текущего устройства из основной структуры текущего устройства
	dev_ex = pci_get_drvdata(dev);

	// Удаление устройства из класса (для удаления узла устройства из директории /dev/)
	device_destroy(nmx_class, MKDEV(Major, dev_ex->Num));

	// Удаление символьного устройства из системы
	cdev_del(&dev_ex->cdev);

	// Деинициализация платы
	nmx_deinit_device(dev_ex);

	// Запрет планирования тасклета
	tasklet_kill(&dev_ex->Interrupt_tasklet);

	// Освобождение памяти из под структуры расширения текущего устройства
	kfree(dev_ex);

	// Отключение устройства
	pci_disable_device(dev);
}

// Инициализация платы
static int nmx_init_device(struct nmx_device_extension *dev_ex)
{
	struct pci_dev *dev; // Указатель на основную структуру текущего устройства
	unsigned int index = dev_ex->Num; // Номер платы
	int Return_value;

	dev = dev_ex->dev; // Указатель на основную структуру текущего устройства

	sprintf(dev_ex->Name, "%s_%2.2u", DEVICE_NAME, index); // Название платы (mcXXXX_XX)

	// Инициализация мьютекса
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
	sema_init(&dev_ex->sem, 1);
#else
	init_MUTEX(&dev_ex->sem);
#endif

	// Получение адресов (не мапированных) и длин ресурсов памяти
	dev_ex->PCICRegs_u = pci_resource_start(dev, 0); // Получение адреса 0-го региона В/В PCI (RUK)
	dev_ex->ShMem_u = pci_resource_start(dev, 2); // Получение адреса 2-го региона В/В PCI (разделяемая память)
	dev_ex->PCICRegs_size = pci_resource_len(dev, 0); // Получение размера 0-го региона В/В PCI (RUK)
	dev_ex->ShMem_size = pci_resource_len(dev, 2); // Получение размера 2-го региона В/В PCI (разделяемая память)

	// Выделение памяти В/В (для RUK)
	if (!request_mem_region(dev_ex->PCICRegs_u, dev_ex->PCICRegs_size, dev_ex->Name))
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Can not get I/O memory address for RUK!\n", index);
		return -ENODEV;
	}

	// Выделение памяти В/В (для разделяемой памяти)
	if (!request_mem_region(dev_ex->ShMem_u, dev_ex->ShMem_size, dev_ex->Name))
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Can not get I/O memory address for shared memory!\n", index);

		// Освобождение памяти В/В (для RUK)
		release_mem_region(dev_ex->PCICRegs_u, dev_ex->PCICRegs_size);

		return -ENODEV;
	}

	// Присвоение виртуальных адресов памяти В/В
	dev_ex->PCICRegs_m = (unsigned long)ioremap_nocache(dev_ex->PCICRegs_u, dev_ex->PCICRegs_size); // RUK
	dev_ex->ShMem_m = (unsigned long)ioremap(dev_ex->ShMem_u, dev_ex->ShMem_size); // Разделяемая память

	dev_ex->irq = dev->irq; // Номер прерывания

	// Установка обработчика прерывания
	Return_value = request_irq(
		dev_ex->irq, // Номер запрашиваемого прерывания
		nmx_irq_handler, // Обработчик
		IRQF_SHARED, // Флаг: Прерывание может быть разделено между устройствами
		dev_ex->Name, // Владелец прерывания
		dev_ex); // Указатель на структуру расширения устройства

	// Если обработчик прерываний не установился
	if (Return_value)
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: IRQ request failed!\n", index);

		// Освобождение виртуальных адресов
		iounmap((void *)dev_ex->PCICRegs_m); // RUK
		iounmap((void *)dev_ex->ShMem_m); // Разделяемая память

		// Освобождение памяти В/В
		release_mem_region(dev_ex->PCICRegs_u, dev_ex->PCICRegs_size); // RUK
		release_mem_region(dev_ex->ShMem_u, dev_ex->ShMem_size); // Разделяемая память

		return Return_value;
	}

	return SUCCESS;
}

// Деинициализация платы
static void nmx_deinit_device(struct nmx_device_extension *dev_ex)
{
	// Освобождение прерывания
	free_irq(dev_ex->irq, dev_ex);

	// Освобождение виртуальных адресов
	iounmap((void *)dev_ex->PCICRegs_m); // RUK
	iounmap((void *)dev_ex->ShMem_m); // Разделяемая память

	// Освобождение памяти В/В
	release_mem_region(dev_ex->PCICRegs_u, dev_ex->PCICRegs_size); // RUK
	release_mem_region(dev_ex->ShMem_u, dev_ex->ShMem_size); // Разделяемая память
}

// Открытие файла устройства
static int nmx_open(struct inode *inode, struct file *filp)
{
	struct nmx_device_extension *dev_ex;

	// Получение структуры расширения устройства
	dev_ex = container_of(inode->i_cdev, struct nmx_device_extension, cdev);
	filp->private_data = dev_ex; // Указатель сохраняется для других методов

	atomic_inc(&dev_ex->Open_count); // Инкремент количества открытий файла устройства

	return SUCCESS;
}

// Закрытие файла устройства
static int nmx_release(struct inode *inode, struct file *filp)
{
	struct nmx_device_extension *dev_ex;

	// Получение структуры расширения устройства
	dev_ex = container_of(inode->i_cdev, struct nmx_device_extension, cdev);

	atomic_dec(&dev_ex->Open_count); // Декремент количества открытий файла устройства

	return SUCCESS;
}

// Изменение текущей позиции в файле
static loff_t nmx_llseek(struct file *filp, loff_t offset, int from)
{
	unsigned int Major_index, Minor_index; // Старший и младший номера устройства
	struct nmx_device_extension *dev_ex; // Указатель на структуру расширения текущего устройства

	// Получение старшего и младшего номеров устройства
	Major_index = imajor(filp->f_dentry->d_inode);
	Minor_index = iminor(filp->f_dentry->d_inode);

	// Получение указателя на структуру расширения текущего устройства
	dev_ex = filp->private_data;

	// Установка новой текущей позиции в файле
	switch (from)
	{
		case SEEK_SET: // Относительно начала файла
			if (offset > dev_ex->ShMem_size) // Если новая позиция находится за концом
			{
				filp->f_pos = dev_ex->ShMem_size;
				break;
			}
			if (offset < 0) // Если новая позиция находится перед началом
			{
				filp->f_pos = 0;
				break;
			}
			filp->f_pos = offset;
			break;

		case SEEK_CUR: // Относительно текущей позиции в файле
			if ((filp->f_pos + offset) > dev_ex->ShMem_size) // Если новая позиция находится за концом
			{
				filp->f_pos = dev_ex->ShMem_size;
				break;
			}
			if ((filp->f_pos + offset) < 0) // Если новая позиция находится перед началом
			{
				filp->f_pos = 0;
				break;
			}
			filp->f_pos += offset;
			break;

		case SEEK_END: // Относительно конца файла
			if (offset > 0) // Если новая позиция находится за концом
			{
				filp->f_pos = dev_ex->ShMem_size;
				break;
			}
			if ((dev_ex->ShMem_size + offset) < 0) // Если новая позиция находится перед началом
			{
				filp->f_pos = 0;
				break;
			}
			filp->f_pos = dev_ex->ShMem_size + offset;
			break;

		default:
			printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Seek ERROR!\n", Minor_index);
			return -EINVAL;
	}

	return filp->f_pos;
}

// Чтение памяти платы
static ssize_t nmx_read(struct file *filp, char __user *buff, size_t count, loff_t *offp)
{
	unsigned int Major_index, Minor_index; // Старший и младший номера устройства
	struct nmx_device_extension *dev_ex; // Указатель на структуру расширения текущего устройства
	u32 Real_offset = (*offp >> 2) << 2; // Действительное смещение
	size_t Real_count = (count >> 2) << 2; // Действительный размер передаваемых данных
	int Return_value;

	// Получение старшего и младшего номеров устройства
	Major_index = imajor(filp->f_dentry->d_inode);
	Minor_index = iminor(filp->f_dentry->d_inode);

	// Получение указателя на структуру расширения текущего устройства
	dev_ex = filp->private_data;

	// Попытка захвата мьютекса текущего устройства
	if (down_interruptible(&dev_ex->sem))
	{
		// Если операция была прервана
		printk(KERN_WARNING DEVICE_NAME" (Board: %u): WARNING: Read operation interrupted!\n", Minor_index);
		return -ERESTARTSYS;
	}

	// Проверка границ и размера передаваемого диапазона данных

	// Если смещение (начало) передаваемых данных выходит за пределы разделяемой памяти
	if (Real_offset >= dev_ex->ShMem_size)
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Incorrect read parameter(s)!\n", Minor_index);

		up(&dev_ex->sem); // Освобождение мьютекса текущего устройства

		return 0; // Передаётся 0 байт
	}

	// Если размер передаваемых данных равен 0
	if (!Real_count)
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Incorrect read parameter(s)!\n", Minor_index);

		up(&dev_ex->sem); // Освобождение мьютекса текущего устройства

		return 0; // Передаётся 0 байт
	}

	// Если конец передаваемых данных выходит за пределы разделяемой памяти
	if (Real_offset + Real_count > dev_ex->ShMem_size)
	{
		printk(KERN_WARNING DEVICE_NAME" (Board: %u): WARNING: Incorrect read parameter(s)! Correcting...\n", Minor_index);

		// Количество передаваемых данных уменьшается, чтобы их конец совпал с концом разделяемой памяти
		Real_count = dev_ex->ShMem_size - Real_offset;
	}

	// Проверка указателя пользовательского пространства
	if (!access_ok(VERIFY_WRITE, buff, Real_count))
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: User space pointer is not valid!\n", Minor_index);

		up(&dev_ex->sem); // Освобождение мьютекса текущего устройства

		return -EFAULT;
	}

	// Пересылка через DMA
	Return_value = nmx_DMA(dev_ex, (unsigned long)buff, Real_offset, Real_count, READ);
	if (Return_value) // Если при пересылке через DMA произошла ошибка
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: DMA transfer ERROR!\n", Minor_index);

		up(&dev_ex->sem); // Освобождение мьютекса текущего устройства

		return Return_value;
	}

	*offp += Real_count; // Смещение увеличивается на количество переданных байт

	up(&dev_ex->sem); // Освобождение мьютекса текущего устройства

	return Real_count;
}

// Запись в память платы
static ssize_t nmx_write(struct file *filp, const char __user *buff, size_t count, loff_t *offp)
{
	unsigned int Major_index, Minor_index; // Старший и младший номера устройства
	struct nmx_device_extension *dev_ex; // Указатель на структуру расширения текущего устройства
	u32 Real_offset = (*offp >> 2) << 2; // Действительное смещение
	u32 Real_count = (count >> 2) << 2; // Действительный размер передаваемых данных
	int Return_value;

	// Получение старшего и младшего номеров устройства
	Major_index = imajor(filp->f_dentry->d_inode);
	Minor_index = iminor(filp->f_dentry->d_inode);

	// Получение указателя на структуру расширения текущего устройства
	dev_ex = filp->private_data;

	// Попытка захвата мьютекса текущего устройства
	if (down_interruptible(&dev_ex->sem))
	{
		// Если операция была прервана
		printk(KERN_WARNING DEVICE_NAME" (Board: %u): WARNING: Write operation interrupted!\n", Minor_index);
		return -ERESTARTSYS;
	}

	// Проверка границ и размера передаваемого диапазона данных

	// Если смещение (начало) передаваемых данных выходит за пределы разделяемой памяти
	if (Real_offset >= dev_ex->ShMem_size)
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Incorrect write parameter(s)!\n", Minor_index);

		up(&dev_ex->sem); // Освобождение мьютекса текущего устройства

		return 0; // Передаётся 0 байт
	}

	// Если размер передаваемых данных равен 0
	if (!Real_count)
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Incorrect write parameter(s)!\n", Minor_index);

		up(&dev_ex->sem); // Освобождение мьютекса текущего устройства

		return 0; // Передаётся 0 байт
	}

	// Если конец передаваемых данных выходит за пределы разделяемой памяти
	if (Real_offset + Real_count > dev_ex->ShMem_size)
	{
		printk(KERN_WARNING DEVICE_NAME" (Board: %u): WARNING: Incorrect write parameter(s)! Correcting...\n", Minor_index);

		// Количество передаваемых данных уменьшается, чтобы их конец совпал с концом разделяемой памяти
		Real_count = dev_ex->ShMem_size - Real_offset;
	}

	// Проверка указателя пользовательского пространства
	if (!access_ok(VERIFY_READ, buff, Real_count))
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: User space pointer is not valid!\n", Minor_index);

		up(&dev_ex->sem); // Освобождение мьютекса текущего устройства

		return -EFAULT;
	}

	// Пересылка через DMA
	Return_value = nmx_DMA(dev_ex, (unsigned long)buff, Real_offset, Real_count, WRITE);
	if (Return_value) // Если при пересылке через DMA произошла ошибка
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: DMA transfer ERROR!\n", Minor_index);

		up(&dev_ex->sem); // Освобождение мьютекса текущего устройства

		return Return_value;
	}

	*offp += Real_count; // Смещение увеличивается на количество переданных байт

	up(&dev_ex->sem); // Освобождение мьютекса текущего устройства

	return Real_count;
}

int nmx_DMA(struct nmx_device_extension *dev_ex, unsigned long User_address, u32 Board_address, u32 Count, int rw)
{
	unsigned int npages; // Количество страниц в пользовательском буфере
	struct page **pages; // Указатель на массив указателей на структуры page
	unsigned int i;
	int Return_value;

	// Получение количества страниц в пользовательском буфере
	npages = ((User_address & ~PAGE_MASK) + Count + ~PAGE_MASK) >> PAGE_SHIFT;

	// Выделение памяти для массива указателей на структуры page
	pages = kmalloc(npages * sizeof(*pages), GFP_KERNEL);
	if (!pages)
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Insufficient memory!\n", dev_ex->Num);
		return -ENOMEM;
	}

	// Получение массива указателей на структуры page, описывающие буфер пользовательского пространства
	down_read(&current->mm->mmap_sem);
	Return_value = get_user_pages(current, current->mm, User_address, npages, rw == READ, 0, pages, NULL);
	up_read(&current->mm->mmap_sem);

	// Если количество отображённых страниц не соответствует запрошенному
	if (Return_value != npages)
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Can not map user buffer!\n", dev_ex->Num);

		// Освобождение всех отображённых страниц из страничного кэша
		for (i = 0; i < Return_value; i++)
		{
			page_cache_release(pages[i]);
		}

		// Освобождение памяти из под массива структур page
		kfree(pages);

		return -EIO;
	}

	// Размаскирование прерывания по завершению DMA (RUK 01)
	*((u32 *)dev_ex->PCICRegs_m + 1) = *((u32 *)dev_ex->PCICRegs_m + 1) | 0x100;

	Return_value = nmx_DMA_(dev_ex, pages, npages, User_address, Board_address, Count, rw);

	// Маскирование прерывания по завершению DMA (RUK 01)
	*((u32 *)dev_ex->PCICRegs_m + 1) = *((u32 *)dev_ex->PCICRegs_m + 1) & ~0x100;

	// Освобождение пользовательских страниц
	for (i = 0; i < npages; i++) // Цикл по всем страницам пользовательского буфера
	{
		// Если запрос на чтение (изменяющий содержимое пользовательского буфера) и
		// страница находится не в зарезервированной части карты памяти (которая никогда не выгружается)
		if ((rw == READ) && (!PageReserved(pages[i])))
		{
			// Пометка страницы как изменённой
			SetPageDirty(pages[i]);
		}

		// Освобождение страницы из страничного кэша
		page_cache_release(pages[i]);
	}

	kfree(pages); // Освобождение памяти из под массива указателей на структуры page

	return Return_value;
}

int nmx_DMA_(struct nmx_device_extension *dev_ex, struct page **pages, unsigned int npages,
	unsigned long User_address, u32 Board_address, u32 Count, int rw)
{
	dma_addr_t PCI_address; // Шинный адрес
	unsigned long Offset; // Смещение от начала страницы
	u32 numxfer = 0; // Количество переданных байтов
	u32 xfer; // Количество байтов, пересылаемых в текущей итерации
	u32 nbytes = Count; // Количество байтов, оставшихся для пересылки
	unsigned int nbuf; // Количество буферов DMA
	struct scatterlist *scl; // Указатель на список разборки
	struct hw_scatterlist *hwl; // Указатель на аппаратный список разборки
	unsigned int i;

	// Выделение памяти для списка разборки
	scl = kmalloc(npages * sizeof(struct scatterlist), GFP_KERNEL);
	if (!scl)
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Insufficient memory!\n", dev_ex->Num);
		return -ENOMEM;
	}

	// Выделение памяти для аппаратного списка разборки
	hwl = kmalloc(npages * sizeof(struct hw_scatterlist), GFP_KERNEL);
	if (!hwl)
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Insufficient memory!\n", dev_ex->Num);

		kfree(scl);

		return -ENOMEM;
	}

	// Создание списка разборки

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 18)
	// Инициализация списка разборки
	sg_init_table(scl, npages);
#endif

	for (i = 0; i < npages; i++) // Цикл по всем страницам пользовательского буфера
	{
		// Получение смещения от начала страницы
		Offset = i ? 0 : (User_address & ~PAGE_MASK);

		// Получение количества байтов в странице
		xfer = nbytes; // Это количество оптимистически приравнивается к количеству оставшихся байтов
		if (xfer > (PAGE_SIZE - Offset)) // Если в странице не достаточно места
		{
			// Количество байтов уменьшается до максимально возможного, чтобы поместилось в странице
			xfer = PAGE_SIZE - Offset;
		}

		// Установка соответствующих полей текущего элемента списка разборки
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 18)
		sg_set_page(&scl[i], pages[i], xfer, Offset);
#else
		scl[i].page = pages[i];
		scl[i].length = xfer;
		scl[i].offset = Offset;
#endif

		nbytes -= xfer; // Количество оставшихся байтов
	} // От цикла по всем страницам пользовательского буфера

	// Получение отображения для страниц, указанных в списке разборки
	nbuf = pci_map_sg(dev_ex->dev, scl, npages, (rw == READ) ? DMA_FROM_DEVICE : DMA_TO_DEVICE);
	if (!nbuf)
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Can not map SG list!\n", dev_ex->Num);

		kfree(hwl);
		kfree(scl);

		return -EIO;
	}

	// Получение отображения для аппаратного списка разборки
	PCI_address = dma_map_single(&dev_ex->dev->dev, hwl, npages * sizeof(struct hw_scatterlist), DMA_TO_DEVICE);
	if (!PCI_address)
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Can not map hardware SG list!\n", dev_ex->Num);

		pci_unmap_sg(dev_ex->dev, scl, npages, (rw == READ) ? DMA_FROM_DEVICE : DMA_TO_DEVICE);

		kfree(hwl);
		kfree(scl);

		return -EIO;
	}

	// Передача права собственности на отображение для аппаратного списка разборки процессору
	dma_sync_single_for_cpu(&dev_ex->dev->dev, PCI_address, npages * sizeof(struct hw_scatterlist), DMA_TO_DEVICE);

	// Создание аппаратного списка разборки

	for (i = 0; i < nbuf; i++) // Цикл по всем буферам DMA
	{
		xfer = sg_dma_len(&scl[i]); // Длина буфера

		hwl[i].PCI_address = sg_dma_address(&scl[i]); // Адрес PCI (два младших разряда равны 0) (аналогично RUK 04)
		hwl[i].Board_address = Board_memory_begin + ((Board_address + numxfer) >> 2); // Адрес разделяемой памяти (аналогично RUK 05)
		hwl[i].Size = ((rw == READ) ? 0x0 : 0x80000000) | (xfer >> 2); // Размер пересылки и направляющий бит (аналогично RUK 06)
		hwl[i].Next = (u32)PCI_address + (i + 1) * sizeof(struct hw_scatterlist); // Указатель на следующий элемент

		numxfer += xfer; // Количество байтов, передаваемых в предыдущих элементах
	}

	hwl[nbuf - 1].Next |= 0x2; // Признак последнего элемента списка

	// Передача права собственности на отображение для аппаратного списка разборки устройству
	dma_sync_single_for_device(&dev_ex->dev->dev, PCI_address, npages * sizeof(struct hw_scatterlist), DMA_TO_DEVICE);

	// Программирование регистров для 2-го режима DMA
	*((u32 *)dev_ex->PCICRegs_m + 4) = (u32)PCI_address; // Адрес PCI (два младших разряда равны 0) (RUK 04)

	// Сброс флага прерывания по завершению DMA
	dev_ex->int_dma = 0;

	// Барьер, чтобы все настройки записались до запуска DMA
	wmb();

	// Запуск 2-го режима DMA
	*((u32 *)dev_ex->PCICRegs_m + 7) = 0x2;

	// Ожидание (в состоянии сна) пока DMA не завершится
	if (wait_event_interruptible(dev_ex->queue, dev_ex->int_dma != 0))
	{
		// Если сон был прерван (не завершением DMA)
		printk(KERN_WARNING DEVICE_NAME" (Board: %u): WARNING: DMA operation interrupted!\n", dev_ex->Num);

		// Останов DMA
		*((u32 *)dev_ex->PCICRegs_m + 7) = 0x8;

		// Освобождение отображения для аппаратного списка разборки
		dma_unmap_single(&dev_ex->dev->dev, PCI_address, npages * sizeof(struct hw_scatterlist), DMA_TO_DEVICE);

		// Освобождение отображения для страниц, указанных в списке разборки
		pci_unmap_sg(dev_ex->dev, scl, npages, (rw == READ) ? DMA_FROM_DEVICE : DMA_TO_DEVICE);

		kfree(hwl); // Освобождение памяти из под аппаратного списка разборки
		kfree(scl); // Освобождение памяти из под списка разборки

		return -ERESTARTSYS;
	}

	// Если ошибка DMA при обращении к шине PCI
	if (*((u32 *)dev_ex->PCICRegs_m + 7) & 0x20)
	{
		printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: DMA ERROR by usage PCI bus!\n", dev_ex->Num);

		// Освобождение отображения для аппаратного списка разборки
		dma_unmap_single(&dev_ex->dev->dev, PCI_address, npages * sizeof(struct hw_scatterlist), DMA_TO_DEVICE);

		// Освобождение отображения для страниц, указанных в списке разборки
		pci_unmap_sg(dev_ex->dev, scl, npages, (rw == READ) ? DMA_FROM_DEVICE : DMA_TO_DEVICE);

		kfree(hwl); // Освобождение памяти из под аппаратного списка разборки
		kfree(scl); // Освобождение памяти из под списка разборки

		return -EIO;
	}

	// Освобождение отображения для аппаратного списка разборки
	dma_unmap_single(&dev_ex->dev->dev, PCI_address, npages * sizeof(struct hw_scatterlist), DMA_TO_DEVICE);

	// Освобождение отображения для страниц, указанных в списке разборки
	pci_unmap_sg(dev_ex->dev, scl, npages, (rw == READ) ? DMA_FROM_DEVICE : DMA_TO_DEVICE);

	kfree(hwl); // Освобождение памяти из под аппаратного списка разборки
	kfree(scl); // Освобождение памяти из под списка разборки

	return SUCCESS;
}

// Обработка ioctl
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
static long nmx_ioctl(struct file *filp, unsigned int ioctl_num, unsigned long ioctl_param)
#else
static int nmx_ioctl(struct inode *inode, struct file *filp, unsigned int ioctl_num, unsigned long ioctl_param)
#endif
{
	unsigned int Major_index, Minor_index; // Старший и младший номера устройства
	struct nmx_device_extension *dev_ex; // Указатель на структуру расширения текущего устройства
	int Return_value = SUCCESS;

	// Получение старшего и младшего номеров устройства
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
	Major_index = imajor(filp->f_dentry->d_inode);
	Minor_index = iminor(filp->f_dentry->d_inode);
#else
	Major_index = imajor(inode);
	Minor_index = iminor(inode);
#endif

	// Получение указателя на структуру расширения текущего устройства
	dev_ex = filp->private_data;

	// Попытка захвата мьютекса текущего устройства
	if (down_interruptible(&dev_ex->sem))
	{
		// Если операция была прервана
		printk(KERN_WARNING DEVICE_NAME" (Board: %u): WARNING: IOCTL operation interrupted!\n", Minor_index);
		return -ERESTARTSYS;
	}

	switch (ioctl_num)
	{
		case IOCTL_NMPCI_GET_RUK: // Чтение управляющего регистра (RUK) платы
			{
				u32 RUK_disp = 0, RUK_val = -1;

				// Получение смещения RUK
				Return_value = get_user(RUK_disp, (u32 __user *)ioctl_param);
				if (Return_value)
				{
					printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Can not get RUK displacement!\n", Minor_index);
					break;
				}

				// Если принятое смещение не соответствует ни одному из RUK
				if ((RUK_disp > 0xE) || (RUK_disp == 0x9))
				{
					printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Incorrect RUK displacement!\n", Minor_index);
					Return_value = -EINVAL;
					break;
				}

				// Чтение значения соответствующего RUK
				RUK_val = *((u32 *)dev_ex->PCICRegs_m + RUK_disp);

				// Передача значения RUK
				Return_value = put_user(RUK_val, (u32 __user *)ioctl_param);
				if (Return_value)
				{
					printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Can not put RUK value!\n", Minor_index);
					break;
				}

				break;
			}

		case IOCTL_NMPCI_SET_RUK: // Запись в управляющий регистр (RUK) платы
			{
				u32 RUK_disp_val[2] = {0, 0};

				// Получение смещения и нового значения RUK
				Return_value = copy_from_user(RUK_disp_val, (void __user *)ioctl_param, 8);
				if (Return_value)
				{
					printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Can not get RUK displacement and new value!\n", Minor_index);
					break;
				}

				// Если принятое смещение не соответствует ни одному из RUK
				if ((RUK_disp_val[0] > 0xE) || (RUK_disp_val[0] == 0x9))
				{
					printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Incorrect RUK displacement!\n", Minor_index);
					Return_value = -EINVAL;
					break;
				}

				// Запись в соответствующий RUK принятого значения
				*((u32 *)dev_ex->PCICRegs_m + RUK_disp_val[0]) = RUK_disp_val[1];

				break;
			}

		case IOCTL_NMPCI_READ_MEMORY: // Чтение ячейки памяти платы
			{
				u32 Mem_addr = 0, Mem_val = 0xDeadBeaf;

				// Получение адреса памяти
				Return_value = get_user(Mem_addr, (int __user *)ioctl_param);
				if (Return_value)
				{
					printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Can not get memory address!\n", Minor_index);
					break;
				}

				// Если принятый адрес лежит за пределом памяти
				if (Mem_addr >= (dev_ex->ShMem_size >> 2))
				{
					printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Incorrect memory address!\n", Minor_index);
					Return_value = -EINVAL;
					break;
				}

				// Чтение значения соответствующей ячейки памяти
				Mem_val = *((u32 *)(dev_ex->ShMem_m) + Mem_addr);

				Return_value = put_user(Mem_val, (int __user *)ioctl_param);
				if (Return_value)
				{
					printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Can not put memory value!\n", Minor_index);
					break;
				}

				break;
			}

		case IOCTL_NMPCI_WRITE_MEMORY: // Запись в ячейку памяти платы
			{
				u32 Mem_addr_val[2] = {0, 0};

				// Получение адреса и нового значения памяти
				Return_value = copy_from_user(Mem_addr_val, (void __user *)ioctl_param, 8);
				if (Return_value)
				{
					printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Can not get memory address and new value!\n", Minor_index);
					break;
				}

				// Если принятый адрес лежит за пределом памяти
				if (Mem_addr_val[0] >= (dev_ex->ShMem_size >> 2))
				{
					printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Incorrect memory address!\n", Minor_index);
					Return_value = -EINVAL;
					break;
				}

				// Запись в соответствующую ячейку памяти нового значения
				*((u32 *)(dev_ex->ShMem_m) + Mem_addr_val[0]) = Mem_addr_val[1];

				break;
			}

		case IOCTL_NMPCI_WAIT_INTERRUPT: // Ожидание прерывания
			{
				struct Wait_interrupt *wi; // Структура ожидания прерывания
				u32 Time_out = 0, Proc_mask = 0x0;

				// Получение значения таймаута
				Return_value = get_user(Time_out, (u32 __user *)ioctl_param);
				if (Return_value)
				{
					printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Can not get timeout value for wait interrupt!\n", Minor_index);
					break;
				}

				// Выделение памяти (с обнолением) для структуры ожидания прерывания
				wi = kzalloc(sizeof(struct Wait_interrupt), GFP_KERNEL);
				if (!wi)
				{
					printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Insufficient memory!\n", Minor_index);
					Return_value = -ENOMEM;
					break;
				}

				wi->tgid = current->tgid; // Идентификатор группы потоков

				up(&dev_ex->sem); // Освобождение мьютекса текущего устройства

				// Добавление к списку ожидания прерывания
				// (под защитой спин-блокировки)
				spin_lock_bh(&dev_ex->Wait_interrupt_lock);
				list_add_tail(&wi->list, &dev_ex->Wait_interrupt_list);
				spin_unlock_bh(&dev_ex->Wait_interrupt_lock);

				// Ожидание (в состоянии сна)
				if (Time_out == 0xFFFFFFFF)
				{
					// Без таймаута
					Return_value = wait_event_interruptible(dev_ex->queue, wi->int_pci);
				}
				else
				{
					// По таймауту
					Return_value = wait_event_interruptible_timeout(dev_ex->queue, wi->int_pci, Time_out * HZ / 1000);
				}

				// Удаление из списка ожидания прерывания
				// (под защитой спин-блокировки)
				spin_lock_bh(&dev_ex->Wait_interrupt_lock);
				list_del(&wi->list);
				spin_unlock_bh(&dev_ex->Wait_interrupt_lock);

				// Сохранение маски процессора
				if (wi->int_pci)
				{
					// Если ожидание в состоянии сна завершилось прерыванием
					// или отменой ожидания прерывания
					Proc_mask = wi->Proc_mask;
				}
				else
				{
					// Если ожидание в состоянии сна завершилось по таймауту
					// (или прерыванием сигналом, но это не важно)
					Proc_mask = 0x80000000;
				}

				// Освобождение памяти из под структуры ожидания прерывания
				kfree(wi);

				// Если ожидание в состоянии сна не было прервано сигналом
				if (Return_value != -ERESTARTSYS)
				{
					// Передача маски процессора
					Return_value = put_user(Proc_mask, (u32 __user *)ioctl_param);
				}
				else
				{
					printk(KERN_WARNING DEVICE_NAME" (Board: %u): WARNING: Wait interrupt interrupted!\n", Minor_index);
				}

				return Return_value;
			}

		case IOCTL_NMPCI_CANCEL_WAIT_INTERRUPT: // Отмена ожидания прерывания
			{
				struct list_head *ptr;
				struct Wait_interrupt *wi;

				up(&dev_ex->sem); // Освобождение мьютекса текущего устройства

				spin_lock_bh(&dev_ex->Wait_interrupt_lock);

				// Цикл по всем структурам из списка ожидания прерывания
				list_for_each(ptr, &dev_ex->Wait_interrupt_list)
				{
					// Получение указателя на очередную структуру ожидания прерывания
					wi = list_entry(ptr, struct Wait_interrupt, list);

					// Установка условия пробуждения для каждого потока из группы текущего потока
					if (wi->tgid == current->tgid)
						wi->int_pci = 1;
				}

				spin_unlock_bh(&dev_ex->Wait_interrupt_lock);

				// Пробуждение соответствующих потоков
				wake_up_interruptible(&dev_ex->queue);

				return Return_value;
			}

		default:
			{
				printk(KERN_ERR DEVICE_NAME" (Board: %u): ERROR: Unknown IOCTL!\n", Minor_index);
				Return_value = -ENOTTY;
			}
	}

	up(&dev_ex->sem); // Освобождение мьютекса текущего устройства

	return Return_value;
}

// Обработчик прерывания
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 18)
static irqreturn_t nmx_irq_handler(int irq, void *dev_id)
#else
static irqreturn_t nmx_irq_handler(int irq, void *dev_id, struct pt_regs *sp)
#endif
{
	struct nmx_device_extension *dev_ex = (struct nmx_device_extension *)dev_id;
	u32 RUK_00 = *((u32 *)dev_ex->PCICRegs_m + 0x0);

	// Если прерывание не от данного устройства
	if (!(RUK_00 & 0x1FF))
	{
		return IRQ_NONE;
	}

	// Если прерывание по завершению DMA
	if (RUK_00 & 0x100)
	{
		// Пробуждение процесса, ожидающего завершение DMA
		dev_ex->int_dma = 1; // Условие для пробуждения
		wake_up_interruptible(&dev_ex->queue); // Непосредственно пробуждение

		return IRQ_HANDLED;
	}

	// Если прерывание PCI
	if (RUK_00 & 0x1)
	{
		// Планирование тасклета прерывания PCI
		tasklet_schedule(&dev_ex->Interrupt_tasklet);

		return IRQ_HANDLED;
	}

	return IRQ_HANDLED;
}

// Тасклет прерывания PCI
void nmx_do_tasklet(unsigned long num)
{
	struct nmx_device_extension *dev_ex = (struct nmx_device_extension *)num;
	struct list_head *ptr;
	struct Wait_interrupt *wi;

	spin_lock(&dev_ex->Wait_interrupt_lock);

	// Цикл по всем структурам из списка ожидания прерывания
	list_for_each(ptr, &dev_ex->Wait_interrupt_list)
	{
		// Получение указателя на очередную структуру ожидания прерывания
		wi = list_entry(ptr, struct Wait_interrupt, list);

		// Установка условия пробуждения
		// (В конечном итоге устанавливается для всех потоков, ожидающих прерывание.)
		wi->int_pci = 1;

		wi->Proc_mask = 0x1; // Маска процессора, вызвавшего прерывание
	}

	spin_unlock(&dev_ex->Wait_interrupt_lock);

	// Непосредственно пробуждение всех потоков, ожидающих прерывание
	wake_up_interruptible(&dev_ex->queue);
}

module_init(nmx_init); // Указание функции инициализации
module_exit(nmx_exit); // Указание функции зачистки
